﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using DistributedGA.Core.Domain;
using DistributedGA.Core.Interface;
using System.Timers;

namespace DistributedGA.Core.Implementation {
  public class WcfPeerListManager : IPeerListManager {

    private string serverString = null;

    private PeerInfo myself = null;

    private Timer timer = null; //sends heartbeat to contact-server

    private Object timerLock = new Object();

    private ChannelFactory<IContactService> myChannelFactory;

    private IContactService client;

    private IContactService heartbeatClient;

    private Random rand;

    private double communicationRate; //how many peers are contacted by this peer in percent

    public void Init(PeerInfo source, string contactServerUrl, double communicationRate) {
      serverString = contactServerUrl;
      this.communicationRate = communicationRate;
      myself = source;
      rand = new Random();
     
      //Init ChannelFactory and Clients
      var binding = new NetTcpBinding();
      var endpoint = new EndpointAddress(serverString);
      myChannelFactory = new ChannelFactory<IContactService>(binding, endpoint);
      client = myChannelFactory.CreateChannel();
      heartbeatClient = myChannelFactory.CreateChannel();
      //Register Peer
      client.RegisterPeer(source);
      //Start heartbeat timer
      timer = new Timer(1000); //each 5 minutes
      timer.Elapsed += SendHeartbeatToServer;
      timer.Start();
    }

    public List<PeerInfo> GetPeerList() {
      try {
        var allPeers = client.GetPeerList(myself); //maybe timout exception...
        var peersForMessaging = ChoosePeersForMessaging(ref allPeers);
        return peersForMessaging;
      }
      catch { } //if maybe sending failed (because of connection lost, etc.): just ignore
      return new List<PeerInfo>();
    }

    public void SendLogToServer(string msg) {
      client.MakeLog(myself, msg);
    }

    public void Dispose() {
      timer.Stop();
      timer.Dispose();
      ((IClientChannel)client).Close();
      ((IClientChannel)heartbeatClient).Close();
      myChannelFactory.Close();
      client = null;
      myChannelFactory = null;
    }


    private List<PeerInfo> ChoosePeersForMessaging(ref List<PeerInfo> allPeers) {
      Shuffle<PeerInfo>(allPeers);
      int toTake = Convert.ToInt32(allPeers.Count * communicationRate) + 1;
      if (allPeers.Count > 0 && toTake == 0) {
        toTake = 1;
      }
      return allPeers.Take(toTake).ToList(); ;
    }
    private  void Shuffle<T>(IList<T> list) {
      int n = list.Count;
      while (n > 1) {
        n--;
        int k = rand.Next(n + 1);
        T value = list[k];
        list[k] = list[n];
        list[n] = value;
      }
    }

    private void SendHeartbeatToServer(object sender, ElapsedEventArgs e) {
      lock (timerLock) {
        try {
          heartbeatClient.UpdateHeartbeat(myself);
        }
        catch { } //nothing to do
      }
    }

  }
}
